
#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include "helper_cuda.h"
#include "helper_math.h"

#include <stdio.h>
#include <vector>

#include <stb_image.h>
#include <stb_image_write.h>

#include "Image.cuh"

bool loadImageF4( const char* filename, ImageF4& img )
{
    int comp;
    stbi_uc* img_uc = stbi_load( filename,&img.width, &img.height, &comp, 4 );
    img.elements = new float4[ img.width * img.height ];
    img.stride_in_bytes = img.width * sizeof( float4 );

    uchar4* img_rgba8 = (uchar4*)img_uc;

    for ( int i = 0; i < img.width * img.height; i++ )
    {
        const uchar4& rgba = img_rgba8[ i ];

        img.elements[ i ] = make_float4( rgba.x / 255.0f, rgba.y / 255.0f, rgba.z / 255.0f, rgba.w / 255.0f );
    }

    stbi_image_free( img_uc );

    return img.elements != nullptr;
}

void saveImageF4( const char* filename, const ImageF4& img )
{
    // Save float4 image to a png

    std::vector<uchar4> img_bytes( img.width * img.height );
    uchar4* p_pixel = img_bytes.data();

    for ( int j = 0; j < img.height; j++ )
        for ( int i = 0; i < img.width; i++ )
        {
            float4 imgPixel = img.GetElement( j, i );
            float4 fData = clamp(imgPixel, 0.0f, 1.0f );

            p_pixel->x = (unsigned char)( fData.x * 255.0f );
            p_pixel->y = (unsigned char)( fData.y * 255.0f );
            p_pixel->z = (unsigned char)( fData.z * 255.0f );
            p_pixel->w = (unsigned char)( fData.w * 255.0f );

            ++p_pixel;
        }
    stbi_write_png( filename, img.width, img.height, 4, img_bytes.data(), img.width * sizeof( uchar4 ) );

}


#define FILTER_SIZE 3
#define BLOCK_SIZE 16

__constant__ float filter_weights[ FILTER_SIZE * 2 + 1 ];

#if 0

// Convolution kernels
__global__ void ConvolutionKernel_Row( ImageF4 imgOut, const ImageF4 img )
{
    uint2 idx = make_uint2( blockIdx.x * blockDim.x + threadIdx.x, blockIdx.y * blockDim.y + threadIdx.y );

    if ( idx.x >= img.width || idx.y >= img.height )
        return;

    float4 sum = make_float4( 0.0f );
    for ( int i = -FILTER_SIZE; i <= FILTER_SIZE; i++ )
    {
        int x = clamp( (int)idx.x + i, 0, img.width - 1 );
        sum += img.GetElement( idx.y, x ) * filter_weights[ i + FILTER_SIZE ];
    }
    
    sum.w = 1.0f;

    imgOut.SetElement( idx.y, idx.x, sum );
}

__global__ void ConvolutionKernel_Col( ImageF4 imgOut, const __grid_constant__ ImageF4 img )
{
    uint2 idx = make_uint2( blockIdx.x * blockDim.x + threadIdx.x, blockIdx.y * blockDim.y + threadIdx.y );

    if ( idx.x >= img.width || idx.y >= img.height )
        return;

    float4 sum = make_float4( 0.0f );
    for ( int i = -FILTER_SIZE; i <= FILTER_SIZE; i++ )
    {
        int y = clamp( (int)idx.y + i, 0, img.height - 1 );
        sum += img.GetElement( y, idx.x ) * filter_weights[ i + FILTER_SIZE ];
    }

    sum.w = 1.0f;

    imgOut.SetElement( idx.y, idx.x, sum );
}

#else
// Convolution kernels
__global__ void ConvolutionKernel_Row( ImageF4 imgOut, const ImageF4 img )
{
    uint2 idx = make_uint2( blockIdx.x * blockDim.x + threadIdx.x, blockIdx.y * blockDim.y + threadIdx.y );

    if ( idx.x >= img.width || idx.y >= img.height )
        return;

    __shared__ float4 shared[ BLOCK_SIZE ][ BLOCK_SIZE + 2 * FILTER_SIZE ];

    shared[ threadIdx.y ][ threadIdx.x + FILTER_SIZE ] = img.GetElement( idx.y, idx.x );
    
    // Filling Halo
    if ( threadIdx.x == 0 )
    {
        for ( int i = 1; i <= FILTER_SIZE; i++ )
        {
            int x = clamp( (int)idx.x - i, 0, img.width - 1 );
            shared[ threadIdx.y ][ threadIdx.x + FILTER_SIZE - i ] = img.GetElement( idx.y, x );
        }
    }

    if ( threadIdx.x == blockDim.x - 1 )
    {
        for ( int i = 1; i <= FILTER_SIZE; i++ )
        {
            int x = clamp( (int)idx.x + i, 0, img.width - 1 );
            shared[ threadIdx.y ][ threadIdx.x + FILTER_SIZE + i ] = img.GetElement( idx.y, x );
        }
    }

    __syncthreads();

    float4 sum = make_float4( 0.0f );

    for ( int i = -FILTER_SIZE; i <= FILTER_SIZE; i++ )
        sum += shared[ threadIdx.y ][ threadIdx.x + FILTER_SIZE + i] * filter_weights[ i + FILTER_SIZE ];

    imgOut.SetElement( idx.y, idx.x, sum );
}

__global__ void ConvolutionKernel_Col( ImageF4 imgOut, const ImageF4 img )
{
    uint2 idx = make_uint2( blockIdx.x * blockDim.x + threadIdx.x, blockIdx.y * blockDim.y + threadIdx.y );

    if ( idx.x >= img.width || idx.y >= img.height )
        return;

    __shared__ float4 shared[ BLOCK_SIZE + 2 * FILTER_SIZE ][ BLOCK_SIZE ];

    shared[ threadIdx.y + FILTER_SIZE ][ threadIdx.x] = img.GetElement( idx.y, idx.x );

    // Filling Halo
    if ( threadIdx.y == 0 )
    {
        for ( int i = 1; i <= FILTER_SIZE; i++ )
        {
            int y = clamp( (int)idx.y - i, 0, img.height - 1 );
            shared[ threadIdx.y + FILTER_SIZE - i  ][ threadIdx.x] = img.GetElement( y, idx.x );
        }
    }

    if ( threadIdx.y == blockDim.y - 1 )
    {
        for ( int i = 1; i <= FILTER_SIZE; i++ )
        {
            int y = clamp( (int)idx.y - i, 0, img.height - 1 );
            shared[ threadIdx.y + FILTER_SIZE + i  ][ threadIdx.x] = img.GetElement( y, idx.x );
        }
    }

    __syncthreads();

    float4 sum = make_float4( 0.0f );

    for ( int i = -FILTER_SIZE; i <= FILTER_SIZE; i++ )
        sum += shared[ threadIdx.y + FILTER_SIZE + i ][ threadIdx.x] * filter_weights[ i + FILTER_SIZE ];

    imgOut.SetElement( idx.y, idx.x, sum );
}
#endif

int main()
{
    // Load image
    ImageF4 img;

    if ( ! loadImageF4("abstract.png", img ) ) {
        fprintf( stderr, "Failed to load image: %s\n", "abstract.png");
        return false;
    }
    
    ImageF4 d_img;
    d_img.width = img.width;
    d_img.height = img.height;
    checkCudaErrors( cudaMallocPitch( &d_img.elements, &d_img.stride_in_bytes, d_img.width * sizeof( float4 ), d_img.height ) );
    checkCudaErrors( cudaMemcpy2D( d_img.elements, d_img.stride_in_bytes, img.elements, img.stride_in_bytes, img.width * sizeof( float4 ), img.height, cudaMemcpyHostToDevice ) );

    ImageF4 d_imgOut_Row;
    d_imgOut_Row.width = img.width;
    d_imgOut_Row.height = img.height;
    checkCudaErrors( cudaMallocPitch( &d_imgOut_Row.elements, &d_imgOut_Row.stride_in_bytes, d_imgOut_Row.width * sizeof( float4 ), d_imgOut_Row.height ) );

    ImageF4 d_imgOut;
    d_imgOut.width = img.width;
    d_imgOut.height = img.height;
    checkCudaErrors( cudaMallocPitch( &d_imgOut.elements, &d_imgOut.stride_in_bytes, d_imgOut.width * sizeof( float4 ), d_imgOut.height ) );

    float filter_weights_h[ FILTER_SIZE * 2 + 1 ];
    for ( int i = 0; i < FILTER_SIZE * 2 + 1; i++ ) filter_weights_h[ i ] = 1.0f / ( FILTER_SIZE * 2 + 1 );

    checkCudaErrors( cudaMemcpyToSymbol( filter_weights, filter_weights_h, sizeof( filter_weights_h ) ) );

    // Convolution kernel

    dim3 blockDim( BLOCK_SIZE,BLOCK_SIZE );
    dim3 gridDim( ( img.width + blockDim.x - 1 ) / blockDim.x, ( img.height + blockDim.y - 1 ) / blockDim.y );

    ConvolutionKernel_Row <<<gridDim, blockDim >>> ( d_imgOut_Row, d_img );
    ConvolutionKernel_Col <<<gridDim, blockDim >>> ( d_imgOut, d_imgOut_Row );

    checkCudaErrors(cudaGetLastError());

    // Copy result back to host
    checkCudaErrors( cudaMemcpy2D( img.elements, img.stride_in_bytes, d_imgOut.elements, d_imgOut.stride_in_bytes, img.width * sizeof( float4 ), img.height, cudaMemcpyDeviceToHost ) );

    saveImageF4( "abstract_out.png", img );

    checkCudaErrors( cudaFree( d_img.elements ) );
    checkCudaErrors( cudaFree( d_imgOut_Row.elements ) );
    checkCudaErrors( cudaFree( d_imgOut.elements ) );

    // Free image
    delete [] img.elements;

    return EXIT_SUCCESS;
}
